"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const flatten = require("flat");
const set = require("lodash/fp/set");
const core_1 = require("../../core");
const isObject = v => Object.prototype.toString.call(v) === '[object Object]' && !v._bsontype;
const flatSet = (obj, key, value, flat) => flat ? Object.assign({}, obj, { [key]: value }) : set(key, value, obj);
const mapObject = (obj, config, activeField) => {
    if (activeField && !(config.continue && config.continue(obj))) {
        const map = (config.valueMaps && config.valueMaps[activeField]);
        return map === true ? obj : map(obj);
    }
    if (!obj)
        return obj;
    if (Array.isArray(obj))
        return obj.map(o => mapObject(o, config));
    if (isObject(obj)) {
        return Object.keys(obj).reduce((res, k) => flatSet(res, (config.newKeys && config.newKeys[k]) || k, mapObject(obj[k], config, activeField ||
            (config.valueMaps && config.valueMaps[k] ? k : undefined)), config.flat), {});
    }
};
function mongoConnector(collection, newId, fieldDbKeys, fieldMaps) {
    const toDbMaps = {
        base: core_1.keysToObject(Object.keys(fieldMaps), k => (fieldMaps[k] && fieldMaps[k].toDb) || true),
        ignoreValues: core_1.keysToObject(Object.keys(fieldMaps), () => true),
    };
    const toDb = (obj, config = {}) => {
        return mapObject(obj, {
            valueMaps: config.ignoreValues ? toDbMaps.ignoreValues : toDbMaps.base,
            newKeys: fieldDbKeys,
            flat: config.flat,
            continue: v => isObject(v) && Object.keys(v).some(k => k[0] === '$'),
        });
    };
    const reverseFieldDbKeys = core_1.keysToObject(Object.keys(fieldDbKeys), k => k, k => fieldDbKeys[k]);
    const fromDb = (doc) => {
        if (!doc)
            return doc;
        const flat = flatten(doc, { safe: true });
        return core_1.keysToObject(Object.keys(flat).map(k => ({
            k: reverseFieldDbKeys[k] || k,
            dbKey: k,
        })), ({ k, dbKey }) => fieldMaps[k]
            ? core_1.mapArray(flat[dbKey], fieldMaps[k].fromDb)
            : flat[dbKey], ({ k }) => k);
    };
    const ops = {
        '=': '$eq',
        '!=': '$ne',
        '<': '$lt',
        '<=': '$lte',
        '>': '$gt',
        '>=': '$gte',
        in: '$in',
    };
    const mongoFilter = (filter) => {
        if (filter[0] === 'AND' || filter[0] === 'OR') {
            return {
                [filter[0] === 'AND' ? '$and' : '$or']: filter[1].map(mongoFilter),
            };
        }
        return {
            [fieldDbKeys[filter[0]] || filter[0]]: { [ops[filter[1]]]: filter[2] },
        };
    };
    return {
        newId,
        async query({ filter, sort, start = 0, end, fields }) {
            if (start === end)
                return [];
            const result = collection.find(filter && toDb(mongoFilter(filter), { flat: true }), toDb(core_1.keysToObject(fields || [], () => true), {
                flat: true,
                ignoreValues: true,
            }), start, core_1.undefOr(end, end - start));
            if (sort) {
                result.sort(toDb(sort.map(([field, order]) => [fieldDbKeys[field] || field, order]), { flat: true, ignoreValues: true }));
            }
            return (await result.toArray()).map(fromDb);
        },
        async findById(id) {
            return fromDb(await collection.findOne({ [fieldDbKeys.id || 'id']: id }));
        },
        async insert(id, data) {
            const obj = Object.assign({ id }, data);
            await collection.insert(toDb(obj));
        },
        async update(id, data) {
            const filter = toDb({ id }, { flat: true });
            await collection.update(filter, { $set: toDb(data, { flat: true }) });
        },
        async delete(id) {
            const filter = toDb({ id }, { flat: true });
            await collection.deleteOne(filter);
        },
        async dump() {
            return (await collection.find().toArray()).map(fromDb);
        },
        async restore(data) {
            try {
                await collection.drop();
            }
            catch (error) { }
            if (data.length > 0) {
                await collection.insertMany(data.map(toDb));
            }
        },
    };
}
exports.default = mongoConnector;
