"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("../../core");
const sqlScalars = {
    boolean: 'BOOLEAN',
    int: 'INTEGER',
    float: 'FLOAT',
    string: 'TEXT',
    date: 'TIMESTAMPTZ',
    file: 'TEXT',
    json: 'JSON',
};
function applyFilter(knex, filter, isOr) {
    if (filter[0] === 'AND' || filter[0] === 'OR') {
        return knex.where(function () {
            filter[1].forEach(f => applyFilter(this, f, filter[0] === 'OR'));
        });
    }
    if (filter[2] === null && ['=', '!='].includes(filter[1])) {
        return knex[`${isOr ? 'orWhere' : 'where'}${filter[1] === '=' ? 'Null' : 'NotNull'}`](filter[0]);
    }
    return knex[isOr ? 'orWhere' : 'where'](filter[0], filter[1], filter[2]);
}
exports.applyFilter = applyFilter;
async function sql(knex, type, newId, fieldTypes, owner) {
    const dbFields = Object.assign({ createdat: 'TIMESTAMPTZ', modifiedat: 'TIMESTAMPTZ' }, core_1.keysToObject(Object.keys(fieldTypes).filter(f => !core_1.fieldIs.foreignRelation(fieldTypes[f])), f => {
        const field = fieldTypes[f];
        return `${sqlScalars[core_1.fieldIs.scalar(field) ? field.scalar : 'string']}${field.isList ? '[]' : ''}`;
    }));
    const columns = (await knex(type).columnInfo());
    if (Object.keys(columns).length === 0) {
        await knex.schema.createTable(type, table => {
            table.text('id').primary();
        });
        if (owner)
            await knex.raw('ALTER TABLE ?? OWNER TO ??;', [type, owner]);
    }
    delete columns.id;
    for (const field of Array.from(new Set([...Object.keys(columns), ...Object.keys(dbFields)]))) {
        if (!columns[field] && dbFields[field]) {
            await knex.schema.table(type, table => {
                table.specificType(field, dbFields[field]);
            });
        }
        else if (columns[field] && !dbFields[field]) {
            // await knex.schema.table(type, table => {
            //   table.dropColumn(field);
            // });
        }
    }
    return {
        newId,
        async query({ filter, sort, start = 0, end, fields }) {
            if (start === end)
                return [];
            const query = filter ? applyFilter(knex(type), filter) : knex(type);
            if (sort) {
                sort.forEach(([field, dir]) => {
                    if (dbFields[field] === 'TEXT') {
                        query.orderByRaw(`lower("${field}") ${dir}`);
                    }
                    else {
                        query.orderByRaw(`${field} ${dir} NULLS LAST`);
                    }
                });
            }
            query.offset(start);
            if (end !== undefined)
                query.limit(end);
            query.select(...(fields || []));
            return await query;
        },
        async findById(id) {
            return await knex(type)
                .where('id', id)
                .first();
        },
        async insert(id, data) {
            await knex(type).insert(Object.assign({ id }, data));
        },
        async update(id, data) {
            await knex(type)
                .where('id', id)
                .update(data);
        },
        async delete(id) {
            await knex(type)
                .where('id', id)
                .delete();
        },
        async dump() {
            return await knex(type).select();
        },
        async restore(data) {
            await knex(type).truncate();
            await knex(type).insert(data);
        },
    };
}
exports.default = sql;
