"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const connectors = require("./connectors");
exports.connectors = connectors;
const graphql_1 = require("graphql");
const core_1 = require("../core");
const batch_1 = require("./batch");
const commit_1 = require("./commit");
const normalize_1 = require("./normalize");
const argTypes = {
    filter: { type: core_1.scalars.json.type },
    sort: { type: core_1.scalars.json.type },
    start: { type: graphql_1.GraphQLInt },
    end: { type: graphql_1.GraphQLInt },
    offset: { type: graphql_1.GraphQLInt },
    trace: {
        type: new graphql_1.GraphQLInputObjectType({
            name: 'Trace',
            fields: {
                start: { type: new graphql_1.GraphQLNonNull(graphql_1.GraphQLInt) },
                end: { type: graphql_1.GraphQLInt },
            },
        }),
    },
};
async function buildServer(schema, options = {}) {
    async function api(request, authHeader) {
        const typeNames = Object.keys(schema);
        const typeFields = core_1.keysToObject(typeNames, type => (Object.assign({ id: { scalar: 'string' }, createdat: { scalar: 'date' }, modifiedat: { scalar: 'date' } }, schema[type].fields)));
        const typeConnectors = core_1.keysToObject(typeNames, type => schema[type].connector);
        const getRecords = async (field, args, { user, internal }, info, extra) => {
            const parsedArgs = core_1.parseArgs(args, user && user.id, typeFields[field.type], core_1.fieldIs.relation(field));
            if (extra) {
                parsedArgs.filter = parsedArgs.filter
                    ? ['AND', [parsedArgs.filter, extra.filter]]
                    : extra.filter;
            }
            parsedArgs.fields = Array.from(new Set([
                'id',
                ...info.fieldNodes[0].selectionSet.selections
                    .map((f) => f.name.value)
                    .filter(fieldName => !core_1.fieldIs.foreignRelation(typeFields[field.type][fieldName])),
                ...(parsedArgs.sort || []).map(([f]) => f),
                ...(extra ? extra.fields : []),
            ]));
            if (internal || !options.auth) {
                return await typeConnectors[field.type].query(parsedArgs);
            }
            const limits = await options.auth.limitQuery(typeFields, runQuery, user, field.type);
            const limitsMap = {};
            limits.forEach(({ filter, fields }) => {
                const key = (fields || []).sort().join('-');
                limitsMap[key] = limitsMap[key] || [];
                if (filter)
                    limitsMap[key].push(filter);
            });
            const groupedLimits = Object.keys(limitsMap).map(key => {
                const fields = key
                    ? ['id', 'createdat', 'modifiedat', ...key.split('-')]
                    : undefined;
                return {
                    filter: parsedArgs.filter && limitsMap[key].length > 0
                        ? ['AND', [parsedArgs.filter, ['OR', limitsMap[key]]]]
                        : parsedArgs.filter || limitsMap[key],
                    fields: parsedArgs.fields && fields
                        ? parsedArgs.fields.filter(f => fields.includes(f))
                        : parsedArgs.fields || fields,
                };
            });
            if (groupedLimits.length === 1) {
                return await typeConnectors[field.type].query(Object.assign({}, parsedArgs, groupedLimits[0]));
            }
            const data = {};
            for (const records of await Promise.all(groupedLimits.map(typeConnectors[field.type].query))) {
                records.forEach(r => (data[r.id] = Object.assign({}, (data[r.id] || {}), r)));
            }
            return Object.keys(data)
                .map(id => data[id])
                .sort(core_1.createCompare((record, key) => record[key], parsedArgs.sort))
                .slice(parsedArgs.start, parsedArgs.end);
        };
        const queryTypes = core_1.keysToObject(typeNames, type => new graphql_1.GraphQLObjectType({
            name: type,
            fields: () => core_1.keysToObject(Object.keys(typeFields[type]), f => {
                const field = typeFields[type][f];
                if (core_1.fieldIs.scalar(field)) {
                    const scalar = f === 'id'
                        ? new graphql_1.GraphQLNonNull(graphql_1.GraphQLID)
                        : core_1.scalars[field.scalar].type;
                    return {
                        type: field.isList
                            ? new graphql_1.GraphQLNonNull(new graphql_1.GraphQLList(scalar))
                            : scalar,
                        description: JSON.stringify(options.auth &&
                            type === options.auth.type &&
                            f === options.auth.usernameField
                            ? Object.assign({}, field, { scalar: 'auth' }) : field),
                        resolve(root) {
                            return root && (field.isList ? root[f] || [] : root[f]);
                        },
                    };
                }
                const relQueryType = queryTypes[field.type];
                return {
                    type: core_1.fieldIs.foreignRelation(field) || field.isList
                        ? new graphql_1.GraphQLNonNull(new graphql_1.GraphQLList(relQueryType))
                        : relQueryType,
                    description: JSON.stringify(field),
                    args: core_1.fieldIs.relation(field) && !field.isList
                        ? undefined
                        : argTypes,
                    resolve: batch_1.default(async (roots, args, context, info) => {
                        const rootField = core_1.fieldIs.relation(field) ? f : 'id';
                        const relField = core_1.fieldIs.relation(field)
                            ? 'id'
                            : field.foreign;
                        const records = await getRecords(field, Object.assign({}, args, { start: undefined, end: undefined }), context, info, {
                            filter: [
                                relField,
                                'in',
                                roots.reduce((res, root) => res.concat(root[rootField] || []), []),
                            ],
                            fields: [relField],
                        });
                        return roots.map(root => {
                            if (core_1.fieldIs.relation(field)) {
                                if (!field.isList) {
                                    return records.find(r => r.id === root[f]);
                                }
                                if (!root[f])
                                    return [];
                                if (args.sort) {
                                    return records
                                        .filter(r => root[f].includes(r.id))
                                        .slice(args.start, args.end);
                                }
                                return root[f]
                                    .slice(args.start, args.end)
                                    .map(id => records.find(r => r.id === id));
                            }
                            return records
                                .filter(r => Array.isArray(r[relField])
                                ? r[relField].includes(root.id)
                                : r[relField] === root.id)
                                .slice(args.start, args.end);
                        });
                    }),
                };
            }),
        }));
        const inputTypes = core_1.keysToObject(typeNames, type => new graphql_1.GraphQLInputObjectType({
            name: `${type}Input`,
            fields: core_1.keysToObject(Object.keys(typeFields[type]), f => {
                const field = typeFields[type][f];
                if (core_1.fieldIs.scalar(field)) {
                    const scalar = f === 'id'
                        ? new graphql_1.GraphQLNonNull(graphql_1.GraphQLID)
                        : core_1.scalars[field.scalar].type;
                    return {
                        type: field.isList ? new graphql_1.GraphQLList(scalar) : scalar,
                    };
                }
                if (core_1.fieldIs.relation(field)) {
                    return {
                        type: field.isList ? new graphql_1.GraphQLList(graphql_1.GraphQLID) : graphql_1.GraphQLID,
                    };
                }
            }),
        }));
        const graphQLSchema = new graphql_1.GraphQLSchema({
            query: new graphql_1.GraphQLObjectType({
                name: 'Query',
                fields: core_1.keysToObject(typeNames, type => ({
                    type: new graphql_1.GraphQLList(new graphql_1.GraphQLNonNull(queryTypes[type])),
                    args: argTypes,
                    async resolve(_root, args, conext, info) {
                        return await getRecords({ type, isList: true }, args, conext, info);
                    },
                })),
            }),
            mutation: new graphql_1.GraphQLObjectType({
                name: 'Mutation',
                fields: {
                    commit: {
                        type: new graphql_1.GraphQLObjectType({
                            name: 'MutationResult',
                            fields: core_1.keysToObject(typeNames, type => ({
                                type: new graphql_1.GraphQLNonNull(new graphql_1.GraphQLList(new graphql_1.GraphQLNonNull(queryTypes[type]))),
                            })),
                        }),
                        args: core_1.keysToObject(typeNames, type => ({
                            type: new graphql_1.GraphQLList(new graphql_1.GraphQLNonNull(inputTypes[type])),
                        })),
                        async resolve(_root, args, context) {
                            return commit_1.default(typeFields, typeConnectors, runQuery, args, context, options.auth);
                        },
                    },
                },
            }),
        });
        const runQuery = async (query) => (await graphql_1.execute(graphQLSchema, graphql_1.parse(query), null, {
            user,
            rootQuery: query,
            mutationsInfo: { mutations: {}, newIds: {} },
            internal: true,
        })).data;
        const queries = Array.isArray(request) ? request : [request];
        let user = null;
        if (options.auth && authHeader && authHeader.split(' ')[0] === 'Bearer') {
            const userId = await options.auth.getUserId(authHeader.split(' ')[1]);
            if (userId) {
                user = Object.assign({ id: userId }, ((typeConnectors[options.auth.type] &&
                    (await typeConnectors[options.auth.type].findById(userId))) ||
                    {}));
            }
            else {
                throw new Error('Not authorized');
            }
        }
        const data = {};
        const mutationsInfo = {
            mutations: {},
            newIds: {},
        };
        const results = await Promise.all(queries.map(async ({ query, variables, normalize }) => {
            const queryDoc = graphql_1.parse(query);
            const result = await graphql_1.execute(graphQLSchema, queryDoc, null, { user, rootQuery: query, mutationsInfo }, variables);
            if (!normalize || result.errors)
                return result;
            return normalize_1.default(typeFields, data, queryDoc, argTypes, result.data);
        }));
        if (options.onMutate && Object.keys(mutationsInfo.mutations).length > 0) {
            await options.onMutate(mutationsInfo.mutations);
        }
        const firstNormalize = queries.findIndex(({ normalize }) => !!normalize);
        if (firstNormalize !== -1) {
            results[firstNormalize].data = data;
            results[firstNormalize].newIds = mutationsInfo.newIds;
        }
        return Array.isArray(request) ? results : results[0];
    }
    return api;
}
exports.default = buildServer;
